/*
Copyright 2018 Google Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS-IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include "geometrical_acoustics/acoustic_ray.h"

#include <limits>

namespace vraudio {

// Static member initialization.
const float AcousticRay::kInfinity = std::numeric_limits<float>::infinity();
const float AcousticRay::kRayEpsilon = 1e-4f;

}  // namespace vraudio
