// Copyright (C) 2023 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only

#ifndef QCGCAPTURABLEWINDOWS_P_H
#define QCGCAPTURABLEWINDOWS_P_H

//
//  W A R N I N G
//  -------------
//
// This file is not part of the Qt API.  It exists purely as an
// implementation detail.  This header file may change from version to
// version without notice, or even be removed.
//
// We mean it.
//

#include <QtMultimedia/private/qplatformcapturablewindows_p.h>

QT_BEGIN_NAMESPACE

namespace QFFmpeg {

class QCGCapturableWindows : public QPlatformCapturableWindows
{
public:
    QList<QCapturableWindow> windows() const override;

    bool isWindowValid(const QCapturableWindowPrivate &window) const override;

    q23::expected<QCapturableWindow, QString> fromQWindow(QWindow *) const override;
};

} // namespace QFFmpeg

QT_END_NAMESPACE

#endif // QCGCAPTURABLEWINDOWS_P_H
